/*
 The software in this package is distributed under the GNU General
 Public License version 2 (with a special exception described below).

 A copy of GNU General Public License (GPL) is included in this distribution,
 in the file COPYING.GPL.

 As a special exception, if other files instantiate templates or use macros
 or inline functions from this file, or you compile this file and link it
 with other works to produce a work based on this file, this file
 does not by itself cause the resulting work to be covered
 by the GNU General Public License.

 However the source code for this file must still be made available
 in accordance with section (3) of the GNU General Public License.

 This exception does not invalidate any other reasons why a work based
 on this file might be covered by the GNU General Public License.

 Christian Herdtweck, Intra2net AG 2015
 */

#ifndef DNS_RESOLVER_H
#define DNS_RESOLVER_H

#include "dns/dnsmaster.h"

#include <boost/asio/deadline_timer.hpp>
#include <boost/asio/ip/udp.hpp>
#include <boost/asio/ip/address.hpp>
#include <boost/system/error_code.hpp>
#include <boost/net/network_array.hpp>   // dns_buffer_t

#include "dns/resolverbase.h"
#include "dns/dnsmaster.h"
#include "dns/dnscache.h"


class DnsResolver : public ResolverBase
{
public:
    ~DnsResolver();

// constructor accessible from friend DnsMaster
public:
    friend ResolverItem& DnsMaster::get_resolver_for(
                                                 const std::string &hostname,
                                                 const DnsIpProtocol &protocol);
private:
    DnsResolver(IoServiceItem &io_serv,
                const std::string &hostname,
                const DnsIpProtocol &protocol,
                const DnsCacheItem cache,
                const boost::asio::ip::address &name_server);

// only real public function (called from pingers)
public:
    HostAddress get_next_ip();
    bool have_up_to_date_ip();
    int get_resolved_ip_count();

// implementation of ResolverBase::async_resolve
protected:
    void do_resolve();

// internal helper functions
private:
    void handle_resolve_timeout(const boost::system::error_code &error);
    void handle_dns_result(const boost::system::error_code &error,
                           const std::size_t bytes_transferred);
    void handle_unavailable();
    void handle_ips(const HostAddressVec &ips);
    void handle_cname(const std::string &canonical_name);
    void cname_resolve_callback(const std::string &canonical_name,
                                const bool was_success,
                                const int cname_count);
    void finalize_resolve(const bool success, const int cname_count=0);
    void stop_trying();
    void wait_timer_timeout_handler(const boost::system::error_code &error);

// variables
private:
    boost::asio::ip::udp::socket Socket;
    boost::net::dns_buffer_t ReceiveBuffer;
    DnsIpProtocol Protocol;
    boost::asio::ip::udp::endpoint NameServer;
    boost::asio::deadline_timer ResolveTimeoutTimer;
    boost::asio::deadline_timer PauseBeforeRetryTimer;
    boost::asio::deadline_timer StaleDataLongtermTimer;
    std::size_t NextIpIndex;
    int RetryCount;
    bool IsResolving;
    std::string LogPrefix;
};

#endif
// (created using vim -- the world's best text editor)

