/*
 The software in this package is distributed under the GNU General
 Public License version 2 (with a special exception described below).

 A copy of GNU General Public License (GPL) is included in this distribution,
 in the file COPYING.GPL.

 As a special exception, if other files instantiate templates or use macros
 or inline functions from this file, or you compile this file and link it
 with other works to produce a work based on this file, this file
 does not by itself cause the resulting work to be covered
 by the GNU General Public License.

 However the source code for this file must still be made available
 in accordance with section (3) of the GNU General Public License.

 This exception does not invalidate any other reasons why a work based
 on this file might be covered by the GNU General Public License.
 */

#ifndef PING_ROTATE_H
#define PING_ROTATE_H

#include <stdint.h>

#include <string>

#include <boost/asio.hpp>
#include <boost/function.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/circular_buffer.hpp>

#include "dns/resolverbase.h"
#include "host/pinger.h"
#include "host/pingprotocol.h"

//-----------------------------------------------------------------------------
// PingRotate
//-----------------------------------------------------------------------------

typedef boost::circular_buffer<PingProtocol> CircularProtocolList;

/**
 * @brief This class is a wrapper to the Pingers, and serves to alternate
 * between protocols.
 * Scope: one object per host.
 */
class PingRotate
{
public:
    PingRotate(
            const IoServiceItem io_serv,
            const std::string &network_interface,
            const std::string &destination_address,
            const uint16_t destination_port,
            const int resolved_ip_ttl_threshold,
            const int ping_reply_timeout,
            const PingProtocolList &protocol_list
    );
    virtual ~PingRotate();

    void ping( boost::function<void(bool)> ping_done_callback );

    void stop_pinging();

    void start_resolving_ping_address();
    int get_resolved_ip_count() const;
    bool have_up_to_date_ip() const;

private:
    //
    // Methods
    //

    void set_ping_done_callback( boost::function<void(bool)> ping_done_callback );
    void ping_done_handler( bool ping_success ) const;

    void init_ping_protocol();
    void update_ping_protocol();
    void get_next_ping_protocol();
    bool can_change_ping_protocol() const;

    void update_dns_resolver( PingProtocol current_protocol );

    void try_to_ping();
    void dns_resolve_callback(const bool was_success, const int cname_count);

    //
    // Attributes
    //

    /// The IO service object, which has the loop event
    IoServiceItem IoService;
    /// The network interface name
    std::string NetworkInterfaceName;
    /// The Dns resolver
    ResolverItem Resolver;
    /// The address to ping
    std::string DestinationAddress;
    /// The port to ping at destination host (same port to all protocols in the list)
    uint16_t DestinationPort;
    /// time threshold for address resolution
    int ResolvedIpTtlThreshold;
    /// timeout for ping reply
    const int PingReplyTimeout;
    /// The list of protocols to ping
    CircularProtocolList ProtocolRotate;
    /// Internal boost pinger object
    PingerItem Ping;
    /// The callback function
    boost::function<void(bool)> PingDoneCallback;
    /// a flag whether DNS resolution is finished or currently underway
    bool DnsResolutionFinished;
    /// a flag whether we should ping as soon as dns is ready
    bool WantToPing;
};

//-----------------------------------------------------------------------------
// PingRotateItem
//-----------------------------------------------------------------------------

typedef boost::shared_ptr<PingRotate> PingRotateItem;

#endif // PING_ROTATE_H
